/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
simple_chip::simple_chip(char *prog, char **opt){
  ev      = new Env();
  ev->sc  = new system_config(prog, opt);
  ev->e   = new evaluation_result();
  ev->as  = new architecture_state(ev);
  ev->mem = new memory_system(ev);
  ev->deb = new debug(ev);
  ev->sys = new system_manager(ev);
  p       = new instruction(ev);
}

simple_chip::~simple_chip(){   /** destructor **/
  print_evaluation_result(ev);
}

/************************************************/
int execute_cmovb(instruction *p, 
		  architecture_state *as){
  INST_TYPE ir=cmov_ir_create(p->get_ir()); 
  if(ir){ /** CMOVb inst, see README.txt **/
    p->Fetch(&as->pc, ir);
    p->Slot();
    p->Rename();
    p->Issue();
    p->RegisterRead();
    p->Execute(&as->pc);
    p->Memory();
    p->WriteBack();  
  }
  return 0;
}

/************************************************/
int simple_chip::step(){
  p->Fetch(&ev->as->pc);   /* pipeline stage 0 */
  p->Slot();               /* pipeline stage 1 */
  p->Rename();             /* pipeline stage 2 */
  p->Issue();              /* pipeline stage 3 */
  p->RegisterRead();       /* pipeline stage 4 */
  p->Execute(&ev->as->pc); /* pipeline stage 5 */
  p->Memory();             /* pipeline stage 6 */
  p->WriteBack();
  
  /* split a conditional move,see README.txt */
  execute_cmovb(p, ev->as);
  
  ev->e->retired_inst++;
  house_keeper(ev);
  
  return ev->sys->running;
}
/************************************************/
