/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"


/* convert hexadecimal value into decimal value */
/************************************************/
uint_64t get_hex(char *buf){
  static char hex_chars[] = {
    '0', '1', '2', '3', '4', '5', '6', '7', 
    '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

  uint_64t sum = 0;

  for(int i=0; i<16; i++){
    char t = buf[i];
    if(t==' ' || t=='\n' || t==0) break;
    int num=0;
    for(int j=0; j<16; j++){
      if(hex_chars[j]==t) num=j;
    }
    sum = sum * 16 + num;
  }
  return sum;
}

/************************************************/
void print_register_name(int i){
  static char *alpha_reg_name[] = {
    "v0",  "t0",  "t1", "t2", 
    "t3",  "t4",  "t5", "t6",
    "t7",  "s0",  "s1", "s2",
    "s3",  "s4",  "s5", "fp",
    "a0",  "a1",  "a2", "a3",
    "a4",  "a5",  "t8", "t9",
    "t10", "t11", "ra", "t12",
    "at",  "gp",  "sp", "zero",
    "pc",  "vfp "
  };
  printf("%-5s", alpha_reg_name[i]);
  return;
}

/************************************************/
void print_status(ADDR_TYPE pc, char *buf, 
		  data_t *reg, uint_64t tc){
  for(int i=0; i<32; i++){
    print_register_name(i);
    printf(" r%-2d      0x%-6llx %-8lld\n", 
	   i, reg[i].ld(), reg[i].ld());
  }
  printf("fpcr           0x0      0\n");
  printf("pc             0x%-6llx %-8lld\n",
	 pc, pc);
  printf("vfp            0x0      0\n");
}


/************************************************/
debug::debug(Env *e){
  ev = e;
}

/************************************************/
void debug::debugmode(){
  char t_buf[256];
  char t_temp[256];
  static int stop = 1;

  char *buf;
  char *temp;
  
  static uint_64t stop_count = 0;
  static ADDR_TYPE stop_address = 0;
  ADDR_TYPE address;

  buf = t_buf;
  temp = t_temp;

  if(stop_count==ev->e->retired_inst){ stop=1; }
  if(stop_address==ev->as->pc.ld()){ stop=1; }
  if(stop==0) return;
  
  if(stop==1){
    data_t d;
    INST_TYPE ir;
    ev->mem->ld_nbyte(4, &ev->as->pc, &d);
    ir = d.ld() & 0xffffffff;

    data_t ir_t;
    ev->mem->ld_nbyte(4, &ev->as->pc, &ir_t);
    printf("TC:%lld  pc:%llx  ir:%08llx\n",
	   ev->e->retired_inst, 
	   ev->as->pc.ld(), ir_t.ld());
    
    print_status(ev->as->pc.ld(), NULL,
		 ev->as->r, ev->e->retired_inst);

    printf("(db)");
    fgets(buf,256, stdin);

    while(*buf != (char)NULL){
      if(!strncmp(buf,"quit",4)) exit(1);
      if(!strncmp(buf,"h",1)){ /* help */
	printf(" command\n");
	printf(" c: stop specified tc.\n");
	printf(" s: stop specified address.\n");
	printf(" x: examine memory contents.\n");
	printf("  quit: quit simulate.\n");
      }
      if(*(buf)=='\n') break;

      if(*(buf)=='x'){
	buf++;
	address = get_hex(buf);
	address = (address & ~7llu);

	for(int j=0; j<16; j++){
	  data_t tmp;
	  printf("0x%09llx: ", address);
	  for(int i=0; i<2; i++){
	    data_t adr;
	    adr.st(address);
	    ev->mem->ld_nbyte(8, &adr, &tmp);
	    printf("0x%016llx ", tmp.ld());
	    address +=8;
	  }
	  printf("\n");
	}
      }
      
      if(*(buf)=='c'){ /* stop by count */
	buf++;
	stop_count = atol(buf);
	printf("stop_tc = %lld\n", stop_count);
	stop=0;
	break;
      }

      if(*(buf)=='s'){ /* stop by address */
	buf++;
	stop_address  = get_hex(buf);
	stop_address  = stop_address;
	printf("stop_address = 0x%llx\n", 
	       stop_address);
	stop=0;
	break;
      }
      
      printf("\n(db)");
      fgets(buf,256,stdin);
    }
  }
}
/************************************************/
