/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
instruction::instruction(Env *ev_t){
  ev = ev_t;
}

/************************************************/
INST_TYPE instruction::get_ir(){
  return ir;
}

/************************************************/
int instruction::Fetch(data_t *pc){
  ev->mem->ld_inst(pc, &ir);
  Cpc.init(pc->ld());
  Npc.init(pc->ld() + 4);
  return 0;
}

/** CMOVb case, npc and pc has the same value  **/
/************************************************/
int instruction::Fetch(data_t *pc, INST_TYPE t){
  ir = t;
  Cpc.init(pc->ld());
  Npc.init(pc->ld());
  return 0;
}

/** Decode IR                                  **/
/************************************************/
int instruction::Slot(){ 
  Op  = (ir>>26) &  0x3F;
  RA  = (ir>>21) &  0x1F;
  RB  = (ir>>16) &  0x1F; 
  RC  = (ir    ) &  0x1F;
  WF  = ((Op&MSK2)==0x14 || (Op&MSK2)==0x20);
  LA  = (Op==0x08 || Op==0x09);
  LD  = (Op==0x0a || Op==0x0b || Op==0x0c ||
	 (Op&MSK2)==0x20 || (Op&MSK2)==0x28);
  ST  = (Op==0x0d || Op==0x0e || Op==0x0f || 
	 (Op&MSK2)==0x24 || (Op&MSK2)==0x2c);
  BR  = ((Op&MSK4)==0x30);
  WB  = (LD || (Op&MSK2)==0x08 || Op==0x1a || 
	 Op==0x30 || Op==0x34) ? RA :
        ((Op&MSK3)==0x10 || Op==0x1c) ? RC : 31;
  Af  = (Op==0x15 || Op==0x16 || Op==0x17 ||
	 Op==0x1c ||
	 (Op&MSK2)==0x24 || (Op&MSK3)==0x30);
  Bf  = ((Op&MSK2)==0x14);
  Ai = (Op==0x08 || Op==0x09 || LD);
  Bi = (BR || (Op&MSK2)==0x10 && (ir & BIT12));

  /** For the CMOV Split Code (CMOV1) **/
  CM = 0;
  if(cmov_ir_create(ir)){ 
    CM = 1; /* This is a cmov inst. */
    RB = RC;
    Bi = 0;
  }
  return 0;
}

/************************************************/
int instruction::Rename(){ 
  return 0;
}

/** genarate the immediate data Imm            **/
/************************************************/
int instruction::Issue(){ 
  DATA_TYPE Lit, D16, D21, tmp, d21e, d16e;
  d21e = ((ir & MASK21) | EXTND21) << 2;
  d16e  = (ir & MASK16) | EXTND16;
  
  Lit = (ir>>13) & 0xFF;
  D21 = (ir & BIT20) ? d21e : (ir&MASK21)<<2;
  D16 = (ir & BIT15) ? d16e : (ir&MASK16);
  if(Op==0x09) D16 = (D16 << 16);
  
  tmp = (LA||LD||ST) ? D16 : (BR) ? D21 : Lit;
  Imm.init(tmp);
  return 0;
}

/** copy data to Rav and Rbv                   **/
/************************************************/
int instruction::RegisterRead(){
  architecture_state *as = ev->as;
  Rav = Ai ? Imm : Af ? as->f[RA] : as->r[RA];
  Rbv = Bi ? Imm : Bf ? as->f[RB] : as->r[RB];
  return 0;
}

/** Updata Rcv, Adr and Tpc                    **/
/************************************************/
int instruction::Execute(data_t *Tpc){ 
  /*** Update Rcv ***/
  if(BR || Op==OP_JSR){
    Rcv.init(Cpc.ld()+4);
  }
  else if(!LD){
    ALU(ir, &Rav, &Rbv, &Rcv); 
  }

  /*** Update Adr ***/
  Adr.init(0);
  if(LD || ST){
    ALU(ir, &Imm, &Rbv, &Adr);
  }
  
  /*** Update Npc ***/
  if(Op==OP_JSR){
    Npc = Rbv;
    Npc.st(Npc.ld() & ~3ull);
  }
  if(BR){
    data_t tmp;
    tmp.init(Cpc.ld()+4);
    BRU(ir, &Rav, &Rbv, &tmp, &Npc);
  }
  *Tpc = Npc;

  return 0;
}

/**  store Rav to mem or load to Rcv           **/
/************************************************/
int instruction::Memory(){ 
  if(ST){
    data_t tmp;
    DATA_TYPE mask;
    set_st_data(ir, &Adr, &Rav, &tmp, &mask);
    ev->mem->st_8byte(&Adr, &tmp, mask);
  }
  if(LD){
    data_t tmp;
    ev->mem->ld_8byte(&Adr, &tmp);
    set_ld_data(ir, &Adr, &tmp, &Rcv);
  }
  return 0;
}

/** PAL or copy Rcv into the regfile           **/
/************************************************/
int instruction::WriteBack(){ 
  if(Op==OP_PAL){
    ev->sys->execute_pal(this);
  }

  if(!WF && WB!=31) ev->as->r[WB] = Rcv;
  if( WF && WB!=31) ev->as->f[WB] = Rcv;
  return 0;
}
/************************************************/
