/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
int nbyte(INST_TYPE ir){
  int Op = (ir>>26) & 0x3F;
  int n = (Op==0x0e || Op==0x0a) ? 1 :
    (Op==0x0d || Op==0x0c) ? 2 :
    (Op==0x22 || Op==0x26 || Op==0x28 || Op==0x2a
     || Op==0x2c || Op==0x2e) ? 4 : 8; 
  return n;
}

/************************************************/
void set_st_data(INST_TYPE ir, data_t *adr,
		 data_t *in, data_t *out, 
		 DATA_TYPE *msk){
  DATA_TYPE a = adr->ld();
  int Op = (ir>>26) & 0x3F;
  DATA_TYPE tmp = in->ld();
  if(Op==0x26){ /***** STS *****/
    DATA_TYPE Rav_t;
    uint_32t int_tmp = 0;
    int_tmp = 0;
    int_tmp |= ((tmp >> 29) & 0x3FFFFFFFul);
    int_tmp |= ((tmp >> 32) & 0xc0000000ul);
    memcpy(&Rav_t, &int_tmp, 4);
    tmp = (Rav_t & MASK32);
  }

  int n = nbyte(ir);
  int offset = a & 7;
  if(n==8) *msk = 0;
  else if(n==4){
    *msk = ~(MASK32 << offset*8);
    tmp = (tmp & MASK32) << offset*8;
  }
  else if(n==2){
    *msk = ~(MASK16 << offset*8);
    tmp = (tmp & MASK16) << offset*8;
  }
  else if(n==1){
    *msk = ~(MASK08 << offset*8);
    tmp = (tmp & MASK08) << offset*8;
  }
  out->st(tmp);
}

/************************************************/
void set_ld_data(INST_TYPE ir, data_t *adr,
		 data_t *in, data_t *out){
  DATA_TYPE a = adr->ld();
  DATA_TYPE tmp=in->ld();
  int offset = a & 7;
  int n = nbyte(ir);
  if(n==4) tmp = (tmp >> (offset * 8)) & MASK32;
  if(n==2) tmp = (tmp >> (offset * 8)) & MASK16;
  if(n==1) tmp = (tmp >> (offset * 8)) & MASK08;
  
  int Op = (ir>>26) & 0x3F;
  if(Op==0x28 || Op==0x2a){ /** LDL & LDL_L **/
    DATA_TYPE Rcv_t = tmp;
    if(Rcv_t & BIT31) Rcv_t|=EXTND32;
    out->st(Rcv_t);
  }
  else if(Op==0x22){         /** LDS **/
    DATA_TYPE rt = tmp;
    rt = (tmp >> 31) & 0x1ull;
    rt = (rt << 11) | (map_s(tmp) & 0x7ffull);
    rt = (rt << 23) | (tmp & 0x7FFFFFull);
    rt =  rt << 29;  
    out->st(rt);
  }
  else {
    out->st(tmp);
  }
}

/************************************************/
inline uint_32t MM_TABLE_INDEX(uint_32t addr){
  return (addr >> BLOCK_MASK_BIT);
}

inline uint_32t gen_tag(uint_32t adr){
  return (adr  & 0xffffffc0);
}

/************************************************/
void memory_system::ld_inst(data_t *a, 
			    INST_TYPE *ir){
  data_t d;
  mm->ld_8byte(a, &d);
  if(a->ld()%8==0) *ir =  d.ld()      & MASK32;
  if(a->ld()%8==4) *ir = (d.ld()>>32) & MASK32;
}

/************************************************/
void memory_system::ld_8byte(data_t *a, data_t *d){
  mm->ld_8byte(a, d);
}

void memory_system::st_8byte(data_t *a, data_t *d,
			     DATA_TYPE msk){
  mm->st_8byte(a, d, msk);
}

/************************************************/
memory_system::~memory_system(){}

/************************************************/
memory_system::memory_system(Env *ev_t){
  ev = ev_t;
  mm = new main_memory(ev);

  char *prog_name = ev->sc->program_name;

  /*** Read memory image from the file ***/
  FILE *fp;
  if((fp = fopen(prog_name, "r")) == NULL) {
    fprintf(stderr, "Bad file name: %s\n", 
	    prog_name);
    exit(1);
  }
  char buf[4096];
  fgets(buf, 4096, fp); /** File Format Check **/
  if(strncmp(buf+3, "SimAlpha", 8)){
    fprintf(stderr, "%s: Not a SimAlpha file.\n", 
	    prog_name);
    exit(1);
  }
  while(!feof(fp)){
    fgets(buf, 4096, fp);
    if(*buf=='@'){

      int ost=0; /* find the offset of dat */
      for(int i=0; i<32; i++){
	if(*(buf+i)==' '){ ost=i; break; }
      }

      ADDR_TYPE adr = get_hex(buf+1);
      DATA_TYPE dat = get_hex(buf+ost+1);
      data_t a,d; 
      a.init(adr);
      d.init(dat);
      mm->st_8byte(&a, &d, 0);
    }
  }
  fclose(fp);
}

/************************************************/
void memory_system::ld_nbyte(int n, data_t *a, 
			     data_t *d){
  if(a->ld()%n!=0) 
    printf("*** ld_nbyte %d miss-align.\n", n);
  if(n!=1 && n!=2 && n!=4 && n!=8)
    printf("*** Error in ld_nbyte %d.\n", n);
  
  ld_8byte(a, d); /** Type Conversion **/
  if(n==8) return;

  int offset = a->ld() & 7;
  if(n==4){
    DATA_TYPE dt = (d->ld()>>(offset*8))&MASK32;
    d->st(dt);
  }
  else if(n==2){
    DATA_TYPE dt = (d->ld()>>(offset*8))&MASK16;
    d->st(dt);
  }
  else if(n==1){
    DATA_TYPE dt = (d->ld()>>(offset*8))&MASK08;
    d->st(dt);
  }
}

/************************************************/
void memory_system::st_nbyte(int n, data_t *a, 
			     data_t *d){
  if(a->ld()%n!=0)
    printf("*** st_nbyte %d miss-alig.\n", n);
  if(n!=1 && n!=2 && n!=4 && n!=8)
    printf("*** Error in st_nbyte %d.\n", n);

  int offset = a->ld() & 7;
  DATA_TYPE mask=0;

  if(n==4){
    mask = ~(MASK32 << offset*8);
    DATA_TYPE dt = (d->ld()&MASK32)<<offset*8;
    d->st(dt);
  }
  else if(n==2){
    mask = ~(MASK16 << offset*8);
    DATA_TYPE dt = (d->ld()&MASK16)<<offset*8;
    d->st(dt);
  }
  else if(n==1){
    mask = ~(MASK08 << offset*8);
    DATA_TYPE dt = (d->ld()&MASK08)<<offset*8;
    d->st(dt);
  }

  st_8byte(a, d, mask); /** Type Conversion **/
}

/************************************************/
main_memory::main_memory(Env *ev_t){
  ev = ev_t;
  for(int i=0; i<BLOCK_TABLE_SIZE; i++){
    block_table[i]=NULL;
  }
}

data_t *main_memory::allocblock(data_t *a){
  data_t *ret =new data_t[BLOCK_SIZE/DATA_T_SIZE];
  for(int i=0; i<BLOCK_SIZE/DATA_T_SIZE; i++){
    ret[i].init(0);
  }
  block_table[MM_TABLE_INDEX(a->ld())]=ret;
  if(ret==NULL){
    printf("** Error in allocblock.\n");
    exit(0);
  }
  ev->e->used_memory_block++;
  return ret;
}

void main_memory::ld_8byte(data_t *a, data_t *d){
  ADDR_TYPE adr = a->ld() & ~7ull;
  data_t *ptr = block_table[MM_TABLE_INDEX(adr)];
  int offset = (adr & BLOCK_MASK)/DATA_T_SIZE;
  if(ptr==NULL) ptr=allocblock(a);
  *d = *(ptr + offset); /** COPY **/
}

void main_memory::st_8byte(data_t *a, data_t *d, 
			   DATA_TYPE msk){
  ADDR_TYPE adr = a->ld() & ~7ull;
  data_t *ptr = block_table[MM_TABLE_INDEX(adr)];
  int offset = (adr & BLOCK_MASK)/DATA_T_SIZE;
  if(ptr==NULL) ptr=allocblock(a);
  (ptr + offset)->st( ((ptr + offset)->ld()
		       & msk) | d->ld() );
}
/************************************************/
